/*****************************************************************************/
/*	SSDT Copyright (C) 2003 Paolo Ardoino				     */
/*****************************************************************************/
/*									     */
/* This program is free software; you can redistribute it and/or modify	     */
/* it under the terms of the GNU General Public License as published by	     */
/* the Free Software Foundation; either version 2 of the License, or	     */
/* (at your option) any later version.					     */
/* This program is distributed in the hope that it will be useful,	     */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of	     */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	     */
/* GNU General Public License for more details.				     */
/*									     */
/* You should have received a copy of the GNU General Public License	     */
/* along with this program; if not, write to the Free Software		     */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*****************************************************************************/


#include <openssl/pem.h>
#include <openssl/err.h>
#include <openssl/rsa.h>


int genkey(int size, char*secfile, char*pubfile)
{
  RSA *key=NULL;
  FILE *fp;

  printf("Generating RSA Key pair [%d bits].\n", size);
  if (size < 64) {
    fprintf(stderr, "Error: RSA Key pair size too small.\n"); 
    fprintf(stderr, "size >= 64\n"); 
    return -1;
  }
  if((key = RSA_generate_key(size,3,NULL,NULL)) == NULL) {
    fprintf(stderr,"%s\n",ERR_error_string(ERR_get_error(),NULL));
    return -1;
  }
  if(RSA_check_key(key) < 1) {
    fprintf(stderr,"Error: Failed generating RSA Key.\nRetry.\n");
    return -1;
  }
  fp=fopen(secfile,"w");
  if(PEM_write_RSAPrivateKey(fp,key,NULL,NULL,0,0,NULL) == 0) {
    fprintf(stderr,"Error: Failed writing RSA Private Key.\n");
    return -1;
  }
  fclose(fp);
  printf("Private Key saved in '%s' file.\n", secfile);
  fp=fopen(pubfile,"w");
  if(PEM_write_RSAPublicKey(fp,key) == 0) {
    fprintf(stderr,"Error: Failed writing RSA Public Key.\n");
    return -1;
  }
  fclose(fp);
  printf("Public Key saved in '%s' file.\n", pubfile);
  RSA_free(key);
  return 0;
}


void* readseckey(char *secfile)
{
  FILE *fp=NULL;
  RSA* key=NULL;
  
  if((fp = fopen(secfile,"r")) == NULL) {
    fprintf(stderr,"Error: Cannot locate Private Key file.\n");
    return NULL;
  }
  if((key = PEM_read_RSAPrivateKey(fp,NULL,NULL,NULL)) == NULL) {
     fprintf(stderr,"Error: Failed reading Private Key.\n");
     return NULL;
   }
   fclose(fp);
   if(RSA_check_key(key) == -1) {
     fprintf(stderr,"Error: Problems while reading RSA Private Key in '%s' file.\n", secfile);
     return NULL;
   } else if(RSA_check_key(key) == 0) {
     fprintf(stderr,"Error: Bad RSA Private Key readed in '%s' file.\n", secfile);
     return NULL;
   }
    else {
      return (void *)key;
    }
}


void *readpubkey(char *pubfile)
{
  FILE *fp;
  RSA *key=NULL;
 
  printf("Loading Public Key. ");
  if((fp = fopen(pubfile, "r")) == NULL) {
    fprintf(stderr,"Error: Cannot locate Public Key file.\n");
    return NULL;
  }
  if((key = PEM_read_RSAPublicKey(fp,NULL,NULL,NULL)) == NULL) {
    fprintf(stderr,"Error: failed reading Public Key in '%s' file.\n", pubfile);
    return NULL;
  }
  fclose(fp);
  printf("Loaded [%d bits].\n", RSA_size(key) * 8);
  return (void *)key;
}


long int rsa_encrypt(char *pubfile, unsigned long int plsz, char *filein, unsigned char *ciphdata)
{
  int ks=0;
  unsigned long size=0, len=0, ctr=0, ciphsz=0;
  RSA* key=NULL;
  FILE *fp=NULL;
  char *tmpciph=NULL, *tmpplain=NULL;
 
  if(!(fp = fopen(filein, "r"))) {
    fprintf(stderr, "Error: Cannot locate input file.\n");
    return -1;
  }
  key = (RSA *)readpubkey(pubfile);
  if (!key) {
    return -1;
  }
  ks = RSA_size(key);
  tmpplain = (unsigned char *)malloc(ks * sizeof(unsigned char));
  tmpciph = (unsigned char *)malloc(ks * sizeof(unsigned char));
  srand(time(NULL));
  while(!feof(fp)) {
    if (ciphsz + ks > plsz) {
      fprintf(stderr, "Error: Size of RSA encrypted data exceeded: %ld bytes.\n", ciphsz);
      fclose(fp);
      return -1;
    }
    memset(tmpplain, '\0', ks);
    memset(tmpciph, '\0', ks);
    len = fread(tmpplain, 1, ks - 11, fp);
    if((size = RSA_public_encrypt(len, tmpplain, tmpciph, key, RSA_PKCS1_PADDING)) == -1) {
      fprintf(stderr, "%s\n", ERR_error_string(ERR_get_error(), NULL));
      return -1;
    }
    for (ctr = 0; ctr < size; ciphsz++, ctr++)
      *(ciphdata + ciphsz) = *(tmpciph + ctr);
  }
  fclose(fp);
  free(tmpciph);
  free(tmpplain);
  RSA_free(key);
  return ciphsz;
}
  
  
long int rsa_decrypt(char *secfile, unsigned char *ciphdata, unsigned long int ciphsz,unsigned char *plaindata)
{
  unsigned long int plsz=0, size=0, ks=0, ctr=0;
  RSA* key=NULL;
  char *tmpplain=NULL, *tmpciph=NULL;
 
  key = (RSA *)readseckey(secfile);
  ks = RSA_size(key);
  tmpciph = (unsigned char *)malloc(ks * sizeof(unsigned char));
  tmpplain = (unsigned char *)malloc(ks * sizeof(unsigned char));
  while (*ciphdata != '\0' && plsz < ciphsz) {
    memset(tmpciph, '\0', ks);
    memset(tmpplain, '\0', ks);
    bcopy(ciphdata, tmpciph, ks);
    for (ctr=0; ctr < ks; ctr++)
      ciphdata++;
    if((size = RSA_private_decrypt(ks, tmpciph, tmpplain, key, RSA_PKCS1_PADDING)) == -1) {
    //  fprintf(stderr, "%s\n", ERR_error_string(ERR_get_error(), NULL));
    //  exit(EXIT_FAILURE);
      continue;
    }
    for(ctr = 0; ctr < size; plsz++, ctr++)
      *(plaindata + plsz) = *(tmpplain + ctr);
  }
  free(tmpplain);
  free(tmpciph);
  RSA_free(key);
  return plsz;
}
